/*
 *  GOL - Game of life simulation software
 *  Copyright (C) 1996 Michael Raymond Brown
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Michael Raymond Brown
 *  EMail:  michael@arrakeen.demon.co.uk
 *  WWW:    http://www.arrakeen.demon.co.uk/
 *  Address:  Michael Brown
 *            Kia Ora
 *            Brook Street
 *            Great Bromley
 *            Nr Colchester
 *            Essex
 *            CO7 7HX
 *            England
 */


#ifdef __GNUC__
#pragma implementation
#endif

#include "App.h"
#include "Image.h"
#include <Xm/Label.h>
#include <Xm/MainW.h>

xwApp::xwApp() : fget_rgb(get_Black), destroy(destroy_Black)
   {int    count   ;
    int     argc   = 1;
    char* argv[]   = {"default"};

    XtSetLanguageProc(0, 0, 0);
    shell = XtVaAppInitialize(&app, "Default", (XrmOptionDescRec*)0, (Cardinal)0, &argc, argv, 0, NULL);
    XtVaSetValues(shell
	, XmNx, -1000, XmNwidth, 100
	, XmNy, -1000, XmNheight, 100
	, 0);
    display = XtDisplay(shell);
    screen  = XtScreen( shell);

    mainwin = XtVaCreateManagedWidget("mainwin"
		, xmMainWindowWidgetClass, shell
		, NULL);

    XtRealizeWidget(shell); 

    XtVaCreateManagedWidget("app", xmLabelWidgetClass, mainwin, NULL);

#ifdef HAVE_MITSHM
    noshm=0;
    for (count=1; count<argc; count++)
	{if (strcmp(argv[count], "-noshm")==0)
	    {noshm=1;
	     for (argc--; count<argc; count++)
		    argv[count]=argv[count+1];
	     break;
	}   }
#endif

    setup_rgb();
   }
xwApp::xwApp(const char* app_name, int& argc, char** argv)
   {int count;

    XtSetLanguageProc(0, 0, 0);
    shell   = XtVaAppInitialize(&app, app_name, (XrmOptionDescRec*)0, (Cardinal)0, &argc, argv, 0, NULL);
    XtVaSetValues(shell
	, XmNtitle, app_name
	, XmNx, -1000, XmNwidth, 100
	, XmNy, -1000, XmNheight, 100
	, 0);
    display = XtDisplay(shell);
    screen  = XtScreen( shell);

    mainwin = XtVaCreateManagedWidget("mainwin"
		, xmMainWindowWidgetClass, shell
		, NULL);


    XtRealizeWidget(shell); 

    XtVaCreateManagedWidget("app", xmLabelWidgetClass, mainwin, NULL);

#ifdef HAVE_MITSHM
    noshm=0;
    for (count=1; count<argc; count++)
	{if (strcmp(argv[count], "-noshm")==0)
	    {noshm=1;
	     for (argc--; count<argc; count++)
		    argv[count]=argv[count+1];
	     break;
	}   }
#endif

    setup_rgb();
   }
xwApp::~xwApp()
   {destroy(this);
    destroyWidgets();
    XtDestroyWidget(mainwin);
    XtDestroyWidget(shell  );
   }


void xwApp::start()
   {XtAppMainLoop(app);
   }


#include <iostream.h>

static inline int get_shift(unsigned long mask)
   {int s = 0;
    while (! (mask&1))
	  {mask>>=1;
	   s++;
	  }
    while (mask&1)
	  {mask>>=1;
	   s++;
	  }
    return s;
   }
static inline int bitcount(unsigned long mask)
   {mask = ((mask>> 1) & 0x55555555) + (mask & 0x55555555);
    mask = ((mask>> 2) & 0x33333333) + (mask & 0x33333333);
    mask = ((mask>> 4) & 0x0f0f0f0f) + (mask & 0x0f0f0f0f);
    mask = ((mask>> 8) & 0x00ff00ff) + (mask & 0x00ff00ff);
    mask = ((mask>>16) & 0x0000ffff) + (mask & 0x0000ffff);
    return mask;
   }

int xwApp::get_TrueColor(xwApp* me, int r, int g, int b)
   {return (((r<<24)>>me->d.xwTrueColor.red_up  )&me->d.xwTrueColor.red_mask  )
	  |(((g<<24)>>me->d.xwTrueColor.green_up)&me->d.xwTrueColor.green_mask)
	  |(((b<<24)>>me->d.xwTrueColor.blue_up )&me->d.xwTrueColor.blue_mask );
   }

int xwApp::get_PseudoColor(xwApp* me, int r, int g, int b)
   {int  col = (r<<16)|(g<<8)|b;
    int   li = r^g^b;
    int   ci = me->d.xwPseudoColor.lookup[li];
    colcell* ptr;
    XColor xcol;

    if (ci >= 0) do {
	ptr = me->d.xwPseudoColor.table + ci;
	if (ptr->color == col) return ptr->pixel;
	ci = ptr->next;
    } while (ci != -1);

    xcol.red   = r << 8;
    xcol.green = g << 8;
    xcol.blue  = b << 8;
    xcol.flags = DoRed|DoGreen|DoBlue;
    if (!XAllocColor(me->display, DefaultColormap(me->display, DefaultScreen(me->display)), &xcol))
	return 0;

       {colcell* nt = new colcell[me->d.xwPseudoColor.tsize+1];
	memcpy(nt, me->d.xwPseudoColor.table, me->d.xwPseudoColor.tsize*sizeof(colcell));
	delete [] me->d.xwPseudoColor.table;
	me->d.xwPseudoColor.table = nt;
	me->d.xwPseudoColor.tsize += 1;
       }

    ptr = me->d.xwPseudoColor.table + me->d.xwPseudoColor.tsize - 1;
    ptr->color = col;
    ptr->pixel = xcol.pixel;
    ptr->next  = me->d.xwPseudoColor.lookup[li];
    me->d.xwPseudoColor.lookup[li]=me->d.xwPseudoColor.tsize - 1;
    return xcol.pixel;
   }
void xwApp::destroy_PseudoColor(xwApp* me)
   {delete [] me->d.xwPseudoColor.table;
    delete [] me->d.xwPseudoColor.lookup;
   }


int xwApp::get_Black(xwApp* me, int r, int g, int b)
   {return 0;
   }
void xwApp::destroy_Black(xwApp* me)
   {
   }


void xwApp::setup_rgb()
   {Visual* vi = DefaultVisual(display, DefaultScreen(display));

    fget_rgb = get_Black;
    destroy = destroy_Black;

    switch (vi->c_class)
       {case DirectColor:
	    cerr << "DirectColor\n";
	    break;
	case GrayScale:
	    cerr << "GrayScale\n";
	    break;
	case PseudoColor:
	    cerr << "PseudoColor\n";

	    this->d.xwPseudoColor.table  = new colcell[1];
	    this->d.xwPseudoColor.tsize  = 0;
	    this->d.xwPseudoColor.lookup = new int[256];
	    memset(this->d.xwPseudoColor.lookup, 0xff, 256*sizeof(int));

	    fget_rgb=get_PseudoColor;
	    destroy=destroy_PseudoColor;
	    fget_image=image_PseudoColor;
	    break;
	case StaticColor:
	    cerr << "StaticColor\n";
	    break;
	case StaticGray:
	    cerr << "StaticGray\n";
	    break;
	case TrueColor:
	    cerr << "TrueColor\n";

	    this->d.xwTrueColor.red_up = 32-get_shift(vi->red_mask);
	    this->d.xwTrueColor.red_mask = vi->red_mask;

	    this->d.xwTrueColor.green_up = 32-get_shift(vi->green_mask);
	    this->d.xwTrueColor.green_mask = vi->green_mask;

	    this->d.xwTrueColor.blue_up = 32-get_shift(vi->blue_mask);
	    this->d.xwTrueColor.blue_mask = vi->blue_mask;

	    fget_rgb=get_TrueColor;
	    fget_image=image_TrueColor;
	    break;
       }
   }





template <class T>
class xwImage_app : public xwImage
   {public:
	xwImage_app(xwApp* app, int w, int h)
	   {width  = w;
	    height = h;
	    display = app->getDisplay();
	    CreateImage(app, ZPixmap, DefaultDepth(display, DefaultScreen(display)));
	    scan_width = ximage->bytes_per_line/sizeof(T);
	   }

	void multiPlot(int x, int y, int w, int h, const int* pixels, int scan_w)
	   {T* ptr = (T*)(ximage->data);
	    int my_w;

#define CLIPEMIN()\
	    while (y < 0) {y++; h--; pixels+=scan_w; }\
	    if ((y + h) >= height) h = height-y;\
\
	    while (x < 0) {x++; w--; pixels++; }\
	    if ((x + w) >= width ) w = width -x;\
\
	    if (h<=0 || w<=0) return;

	    CLIPEMIN()
	    ptr += (scan_width*y)+x;

	    scan_w -= w;
	    my_w = scan_width - w;
	    do {int tw = w;
		do {*ptr++=*pixels++;
		} while (--tw);
		pixels+=scan_w;
		ptr+=my_w;
	    } while (--h);
	   }
   };


class xwImage_bmap : public xwImage
   {public:
	xwImage_bmap(xwApp* app, int w, int h)
	   {width  = w;
	    height = h;
	    display = app->getDisplay();
	    CreateImage(app, XYBitmap, 1);
	   }

	void multiPlot(int x, int y, int w, int h, const int* pixels, int scan_w)
	   {int      count;
	    int          W = ximage->bytes_per_line / sizeof(int);
	    u_int* dst_ptr;
	    const int* src_ptr;
	    CLIPEMIN()

	    dst_ptr = ((u_int*)(ximage->data+(ximage->bytes_per_line*y))) + (x/BITSPERINT);
	    src_ptr = pixels;

	    do {
		const int*  pix;
		u_int*  dst = dst_ptr;
		u_int value;
		u_int  mask;
		  int shift = x&15;
		  int  bits = BITSPERINT-shift;
		if (bits>w) bits=w;
		mask = ~(((1U<<bits)-1U)<<shift);

		count = h;
		pix = pixels - shift;
		do {int sh = shift;
		    int bi = bits;
		    value = *dst & mask;

		    while (bi >= 4)
		       {value |= pix[sh]<<sh; sh<<=1;
			value |= pix[sh]<<sh; sh<<=1;
			value |= pix[sh]<<sh; sh<<=1;
			value |= pix[sh]<<sh; sh<<=1;
			bi-=4;
		       }
		    if (bi >= 2)
		       {value |= pix[sh]<<sh; sh<<=1;
			value |= pix[sh]<<sh; sh<<=1;
			bi-=2;
		       }
		    if (bi)
		       {value |= pix[sh]<<sh;
		       }

		    *dst = value;
		    pixels += scan_w;
		    dst += W;
		} while (--count);

		count=BITSPERINT-(x&(BITSPERINT-1));
		src_ptr += count;
		pixels = src_ptr;
		dst_ptr += 1;
		x += count;
		w -= count;
	    } while (w>0);
	   }
   };

xwImage* xwApp::image_TrueColor(xwApp* me, int w, int h)
   {return new xwImage_app<unsigned short>(me, w, h);
   }
xwImage* xwApp::image_PseudoColor(xwApp* me, int w, int h)
   {return new xwImage_app<unsigned char>(me, w, h);
   }


