/*
 *  GOL - Game of life simulation software
 *  Copyright (C) 1996 Michael Raymond Brown
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Michael Raymond Brown
 *  EMail:  michael@arrakeen.demon.co.uk
 *  WWW:    http://www.arrakeen.demon.co.uk/
 *  Address:  Michael Brown
 *            Kia Ora
 *            Brook Street
 *            Great Bromley
 *            Nr Colchester
 *            Essex
 *            CO7 7HX
 *            England
 */


#ifdef __GNUC__
#pragma implementation
#endif

#include "App.h"
#include "Button.h"

void xwButton::init()
   {auto_on    =   0;
    auto_delay = 250;
    auto_speed =  10;
    armed      =   0;
    default_callbacks(0, meth_Caption|meth_Size|meth_Close);
    XtAddCallback(widget, XmNarmCallback, cbArm, (XtPointer)this);
    XtAddCallback(widget, XmNdisarmCallback, cbDisarm, (XtPointer)this);
    linkParent(parent);
   }

xwButton::xwButton(xwWidget* p, const char* name, int x1, int y1, int x2, int y2)
   {parent = p;
    XmString label = XmStringCreateLocalized((char*)name);
    widget = XtVaCreateManagedWidget("button"
		, xmPushButtonWidgetClass, parent->getWidget()
		, XmNlabelString, label

		, XmNleftAttachment,	XmATTACH_POSITION
		, XmNrightAttachment,	XmATTACH_POSITION
		, XmNtopAttachment,	XmATTACH_POSITION
		, XmNbottomAttachment,	XmATTACH_POSITION

		, XmNleftPosition,	x1
		, XmNrightPosition,	x2
		, XmNtopPosition,	y1
		, XmNbottomPosition,	y2

		, XmNalignment, XmALIGNMENT_CENTER

		, 0);
    XmStringFree(label);
    init();
   }
xwButton::xwButton(xwWidget* p, const char* name)
   {parent = p;
    XmString label = XmStringCreateLocalized((char*)name);
    widget = XtVaCreateManagedWidget("button"
		, xmPushButtonWidgetClass, parent->getWidget()
		, XmNlabelString, label
		, XmNalignment, XmALIGNMENT_CENTER
		, 0);
    XmStringFree(label);
    init();
   }
xwButton::~xwButton()
   {if (armed) XtRemoveTimeOut(int_id);
   }


void xwButton::cbArm(Widget w, XtPointer client_data, XtPointer call_data)
   {xwButton* me = (xwButton*)client_data;
    if (me->auto_on)
       {if (me->armed) XtRemoveTimeOut(me->int_id);
	me->armed=1;
	me->int_id=XtAppAddTimeOut(me->getApp()->getApp(),me->auto_delay,cbIdle,(XtPointer)me);
       }
    if (me->handler)
       {if (me->auto_on) me->doClick(call_data);
	me->handler->onArm(me, me->handler_arg);
   }   }
void xwButton::cbDisarm(Widget w, XtPointer client_data, XtPointer call_data)
   {xwButton* me = (xwButton*)client_data;
    if (me->armed)
       {XtRemoveTimeOut(me->int_id);
	me->armed=0;
       }
    if (me->handler)
       {me->handler->onDisarm(me, me->handler_arg);
	if (!me->auto_on) me->doClick(call_data);
   }   }

void xwButton::doClick(XtPointer p)
   {XmPushButtonCallbackStruct* cbs = (XmPushButtonCallbackStruct*)p;
    int x = cbs->event->xbutton.x;
    int y = cbs->event->xbutton.y;
    int w , h;
    getSize(&w, &h);
    w--; h--;
    if (x>0 && y>0 && x<w && y<h)
	handler->onClick(this, handler_arg);
   }

void xwButton::cbIdle(XtPointer client, XtIntervalId* id)
   {xwButton* me = (xwButton*)client;
    me->int_id=XtAppAddTimeOut(me->getApp()->getApp(),1000/me->auto_speed,cbIdle,(XtPointer)me);
    if (me->handler) me->handler->onClick(me, me->handler_arg);
   }


void xwButton::setDelay(int delay)
   {auto_delay = delay;
   }
int xwButton::getDelay()
   {return auto_delay;
   }

void xwButton::setSpeed(int speed)
   {auto_speed = speed;
   }
int xwButton::getSpeed()
   {return auto_speed;
   }

void xwButton::setAuto(int auto_on)
   {this->auto_on = auto_on;
   }
int xwButton::getAuto()
   {return auto_on;
   }

